# Copyright (c) Felix Petersen.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from . import functional
from .mesh import Mesh
from .transform import Projection, LookAt, Look
from .lighting import AmbientLighting, DirectionalLighting, Lighting
from .renderer import GenDR
from .losses import LaplacianLoss, FlattenLoss



# Copyright (c) Felix Petersen.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import torch
import torch.nn.functional as F


def vertex_normals(vertices, faces):
    """
    :param vertices: [batch size, number of vertices, 3]
    :param faces: [batch size, number of faces, 3]
    :return: [batch size, number of vertices, 3]
    """
    assert (vertices.ndimension() == 3)
    assert (faces.ndimension() == 3)
    assert (vertices.shape[0] == faces.shape[0])
    assert (vertices.shape[2] == 3)
    assert (faces.shape[2] == 3)

    bs, nv = vertices.shape[:2]
    bs, nf = faces.shape[:2]
    device = vertices.device
    normals = torch.zeros(bs * nv, 3).to(device)

    faces = faces + (torch.arange(bs, dtype=torch.int32).to(device) * nv)[:, None, None]  # expanded faces
    vertices_faces = vertices.reshape((bs * nv, 3))[faces.long()]

    faces = faces.view(-1, 3)
    vertices_faces = vertices_faces.view(-1, 3, 3)

    normals.index_add_(0, faces[:, 1].long(),
                       torch.cross(vertices_faces[:, 2] - vertices_faces[:, 1],
                                   vertices_faces[:, 0] - vertices_faces[:, 1]))
    normals.index_add_(0, faces[:, 2].long(),
                       torch.cross(vertices_faces[:, 0] - vertices_faces[:, 2],
                                   vertices_faces[:, 1] - vertices_faces[:, 2]))
    normals.index_add_(0, faces[:, 0].long(),
                       torch.cross(vertices_faces[:, 1] - vertices_faces[:, 0],
                                   vertices_faces[:, 2] - vertices_faces[:, 0]))

    normals = F.normalize(normals, eps=1e-6, dim=1)
    normals = normals.reshape((bs, nv, 3))
    # pytorch only supports long and byte tensors for indexing
    return normals



# Copyright (c) Felix Petersen.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import torch

import gendr.cuda.voxelization as voxelization_cuda


def voxelize_sub1(faces, size, dim):
    bs = faces.size(0)
    nf = faces.size(1)
    if dim == 0:
        faces = faces[:, :, :, [2, 1, 0]].contiguous()
    elif dim == 1:
        faces = faces[:, :, :, [0, 2, 1]].contiguous()
    voxels = torch.zeros(bs, size, size, size).int().cuda()
    return voxelization_cuda.voxelize_sub1(faces, voxels)[0].transpose(dim + 1, -1)


def voxelize_sub2(faces, size):
    bs = faces.size(0)
    nf = faces.size(1)
    voxels = torch.zeros(bs, size, size, size).int().cuda()
    return voxelization_cuda.voxelize_sub2(faces, voxels)[0]


def voxelize_sub3(faces, voxels):
    bs = voxels.size(0)
    vs = voxels.size(1)
    visible = torch.zeros_like(voxels, dtype=torch.int32).cuda()
    voxels, visible = voxelization_cuda.voxelize_sub3(faces, voxels, visible)

    sum_visible = visible.sum()

    while True:
        voxels, visible = voxelization_cuda.voxelize_sub4(faces, voxels, visible)
        if visible.sum() == sum_visible:
            break
        else:
            sum_visible = visible.sum()
    return 1 - visible


def voxelization(faces, size, normalize=False):
    faces = faces.clone()
    if normalize:
        pass
    else:
        faces *= size

    voxels0 = voxelize_sub1(faces, size, 0)
    voxels1 = voxelize_sub1(faces, size, 1)
    voxels2 = voxelize_sub1(faces, size, 2)
    voxels3 = voxelize_sub2(faces, size)

    voxels = voxels0 + voxels1 + voxels2 + voxels3
    voxels = (voxels > 0).int().contiguous()
    voxels = voxelize_sub3(faces, voxels)

    return voxels



# Copyright (c) Felix Petersen.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from .get_points_from_angles import get_points_from_angles
from .lighting import ambient_lighting, directional_lighting
from .load_obj import load_obj
from .look import look
from .look_at import look_at
from .renderer import render
from .save_obj import (save_obj, save_voxel)
from .face_vertices import face_vertices
from .vertex_normals import vertex_normals
from .voxelization import voxelization



# Copyright (c) Felix Petersen.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import math
import numpy as np
import torch
import torch.nn as nn

import gendr


def perspective(vertices, angle=30.):
    '''
    Compute perspective distortion from a given angle
    '''
    if (vertices.ndimension() != 3):
        raise ValueError('vertices Tensor should have 3 dimensions')
    device = vertices.device
    angle = torch.tensor(angle / 180 * math.pi, dtype=torch.float32, device=device)
    angle = angle[None]
    width = torch.tan(angle)
    width = width[:, None]
    z = vertices[:, :, 2]
    x = vertices[:, :, 0] / z / width
    y = vertices[:, :, 1] / z / width
    vertices = torch.stack((x, y, z), dim=2)
    return vertices


def orthogonal(vertices, scale=1.):
    '''
    Compute orthogonal projection from a given angle
    To find equivalent scale to perspective projection
    set scale = focal_pixel / object_depth  -- to 0~H/W pixel range
              = 1 / ( object_depth * tan(half_fov_angle) ) -- to -1~1 pixel range
    '''
    if (vertices.ndimension() != 3):
        raise ValueError('vertices Tensor should have 3 dimensions')
    z = vertices[:, :, 2]
    x = vertices[:, :, 0] * scale
    y = vertices[:, :, 1] * scale
    vertices = torch.stack((x, y, z), dim=2)
    return vertices


class Transform(nn.Module):
    def __init__(self):
        super().__init__()

    def transform(self, vertices):
        raise NotImplementedError()

    def forward(self, mesh):
        new_vertices = self.transform(mesh.vertices)
        faces = mesh.faces
        textures = mesh.textures
        texture_res = mesh.texture_res
        texture_type = mesh.texture_type
        return gendr.Mesh(new_vertices, faces, textures, texture_res, texture_type)


class Projection(Transform):
    def __init__(self, P, dist_coeffs=None, orig_size=512):
        super().__init__()
        '''
        Calculate projective transformation of vertices given a projection matrix
        P: 3x4 projection matrix
        dist_coeffs: vector of distortion coefficients
        orig_size: original size of image captured by the camera
        '''

        self.P = P
        self.dist_coeffs = dist_coeffs
        self.orig_size = orig_size

        if isinstance(self.P, np.ndarray):
            self.P = torch.from_numpy(self.P).cuda()
        if self.P is None or self.P.ndimension() != 3 or self.P.shape[1] != 3 or self.P.shape[2] != 4:
            raise ValueError('You need to provide a valid (batch_size)x3x4 projection matrix')
        if dist_coeffs is None:
            self.dist_coeffs = torch.cuda.FloatTensor([[0., 0., 0., 0., 0.]]).repeat(self.P.shape[0], 1)

    def transform(self, vertices):
        vertices = torch.cat([vertices, torch.ones_like(vertices[:, :, None, 0])], dim=-1)
        vertices = torch.bmm(vertices, self.P.transpose(2, 1))
        x, y, z = vertices[:, :, 0], vertices[:, :, 1], vertices[:, :, 2]
        x_ = x / (z + 1e-5)
        y_ = y / (z + 1e-5)

        # Get distortion coefficients from vector
        k1 = self.dist_coeffs[:, None, 0]
        k2 = self.dist_coeffs[:, None, 1]
        p1 = self.dist_coeffs[:, None, 2]
        p2 = self.dist_coeffs[:, None, 3]
        k3 = self.dist_coeffs[:, None, 4]

        # we use x_ for x' and x__ for x'' etc.
        r = torch.sqrt(x_ ** 2 + y_ ** 2)
        x__ = x_*(1 + k1*(r**2) + k2*(r**4) + k3*(r**6)) + 2*p1*x_*y_ + p2*(r**2 + 2*x_**2)
        y__ = y_*(1 + k1*(r**2) + k2*(r**4) + k3 * (r**6)) + p1*(r**2 + 2*y_**2) + 2*p2*x_*y_
        x__ = 2 * (x__ - self.orig_size / 2.) / self.orig_size
        y__ = 2 * (y__ - self.orig_size / 2.) / self.orig_size
        vertices = torch.stack([x__, y__, z], dim=-1)
        return vertices


class LookAt(Transform):
    def __init__(self, perspective=True, viewing_angle=30, viewing_scale=1.0, eye=None):
        super(LookAt, self).__init__()

        self.perspective = perspective
        self.viewing_angle = viewing_angle
        self.viewing_scale = viewing_scale
        self._eye = eye

        if self._eye is None:
            self._eye = [0, 0, -(1. / math.tan(math.radians(self.viewing_angle)) + 1)]

    def set_eyes_from_angles(self, distances, elevations, azimuths):
        self._eye = gendr.functional.get_points_from_angles(distances, elevations, azimuths)

    def set_eyes(self, eyes):
        self._eye = eyes

    @property
    def eyes(self):
        return self._eyes

    def transform(self, vertices):
        vertices = gendr.functional.look_at(vertices, self._eye)
        # perspective transformation
        if self.perspective:
            vertices = perspective(vertices, angle=self.viewing_angle)
        else:
            vertices = orthogonal(vertices, scale=self.viewing_scale)
        return vertices


class Look(Transform):
    def __init__(self, camera_direction=[0, 0, 1], perspective=True, viewing_angle=30, viewing_scale=1.0, eye=None):
        super(Look, self).__init__()

        self.perspective = perspective
        self.viewing_angle = viewing_angle
        self.viewing_scale = viewing_scale
        self._eye = eye
        self.camera_direction = camera_direction

        if self._eye is None:
            self._eye = [0, 0, -(1. / math.tan(math.radians(self.viewing_angle)) + 1)]

    def set_eyes(self, eyes):
        self._eye = eyes

    @property
    def eyes(self):
        return self._eyes

    def transform(self, vertices):
        vertices = gendr.functional.look(vertices, self._eye, self.camera_direction)
        # perspective transformation
        if self.perspective:
            vertices = perspective(vertices, angle=self.viewing_angle)
        else:
            vertices = orthogonal(vertices, scale=self.viewing_scale)
        return vertices



import torch
import torch.nn as nn

# Copyright (c) Felix Petersen.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import gendr


class AmbientLighting(nn.Module):
    def __init__(self, light_intensity=0.5, light_color=(1, 1, 1)):
        super(AmbientLighting, self).__init__()

        self.light_intensity = light_intensity
        self.light_color = light_color

    def forward(self, light):
        return gendr.functional.ambient_lighting(light, self.light_intensity, self.light_color)


class DirectionalLighting(nn.Module):
    def __init__(self, light_intensity=0.5, light_color=(1, 1, 1), light_direction=(0, 1, 0)):
        super(DirectionalLighting, self).__init__()

        self.light_intensity = light_intensity
        self.light_color = light_color
        self.light_direction = light_direction

    def forward(self, light, normals):
        return gendr.functional.directional_lighting(light, normals,
                                        self.light_intensity, self.light_color,
                                        self.light_direction)


class Lighting(nn.Module):
    def __init__(self, intensity_ambient=0.5, color_ambient=[1, 1, 1],
                 intensity_directionals=0.5, color_directionals=[1, 1, 1],
                 directions=[0, 1, 0]):
        super(Lighting, self).__init__()

        self.ambient = AmbientLighting(intensity_ambient, color_ambient)
        self.directionals = nn.ModuleList([DirectionalLighting(intensity_directionals,
                                                               color_directionals,
                                                               directions)])

    def forward(self, mesh):
        if mesh.texture_type == 'surface':
            light = torch.zeros_like(mesh.faces, dtype=torch.float32).to(mesh.device)
            light = light.contiguous()
            light = self.ambient(light)
            for directional in self.directionals:
                light = directional(light, mesh.surface_normals)
            new_textures = mesh.textures * light[:, :, None, :]

        elif mesh.texture_type == 'vertex':
            light = torch.zeros_like(mesh.vertices, dtype=torch.float32).to(mesh.device)
            light = light.contiguous()
            light = self.ambient(light)
            for directional in self.directionals:
                light = directional(light, mesh.vertex_normals)
            new_textures = mesh.textures * light

        vertices = mesh.vertices
        faces = mesh.faces
        texture_res = mesh.texture_res
        texture_type = mesh.texture_type
        mesh = gendr.Mesh(vertices, faces, new_textures, texture_res, texture_type)

        return mesh



# Copyright (c) Felix Petersen.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import numpy as np
import torch
import torch.nn.functional as F


def look_at(vertices, eye, at=[0, 0, 0], up=[0, 1, 0], only_rotate=False):
    """
    "Look at" transformation of vertices.
    """
    if (vertices.ndimension() != 3):
        raise ValueError('vertices Tensor should have 3 dimensions')

    device = vertices.device

    # if list or tuple convert to numpy array
    if isinstance(at, list) or isinstance(at, tuple):
        at = torch.tensor(at, dtype=torch.float32, device=device)
    # if numpy array convert to tensor
    elif isinstance(at, np.ndarray):
        at = torch.from_numpy(at).to(device)
    elif torch.is_tensor(at):
        at.to(device)

    if isinstance(up, list) or isinstance(up, tuple):
        up = torch.tensor(up, dtype=torch.float32, device=device)
    elif isinstance(up, np.ndarray):
        up = torch.from_numpy(up).to(device)
    elif torch.is_tensor(up):
        up.to(device)

    if isinstance(eye, list) or isinstance(eye, tuple):
        eye = torch.tensor(eye, dtype=torch.float32, device=device)
    elif isinstance(eye, np.ndarray):
        eye = torch.from_numpy(eye).to(device)
    elif torch.is_tensor(eye):
        eye = eye.to(device)

    batch_size = vertices.shape[0]
    if eye.ndimension() == 1:
        eye = eye[None, :].repeat(batch_size, 1)
    if at.ndimension() == 1:
        at = at[None, :].repeat(batch_size, 1)
    if up.ndimension() == 1:
        up = up[None, :].repeat(batch_size, 1)

    # create new axes
    # eps is chosen as 1e-5 to match the chainer version
    z_axis = F.normalize(at - eye, eps=1e-5)
    x_axis = F.normalize(torch.cross(up, z_axis, dim=1), eps=1e-5)
    y_axis = F.normalize(torch.cross(z_axis, x_axis, dim=1), eps=1e-5)

    # create rotation matrix: [bs, 3, 3]
    r = torch.cat((x_axis[:, None, :], y_axis[:, None, :], z_axis[:, None, :]), dim=1)

    # apply
    # [bs, nv, 3] -> [bs, nv, 3] -> [bs, nv, 3]
    if vertices.shape != eye.shape:
        eye = eye[:, None, :]
    if not only_rotate:
        vertices = vertices - eye
    vertices = torch.matmul(vertices, r.transpose(1, 2))

    return vertices



# Copyright (c) Felix Petersen.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import torch
import torch.nn as nn
import numpy as np


class LaplacianLoss(nn.Module):
    def __init__(self, vertex, faces, average=False):
        super(LaplacianLoss, self).__init__()
        self.nv = vertex.size(0)
        self.nf = faces.size(0)
        self.average = average
        laplacian = np.zeros([self.nv, self.nv]).astype(np.float32)

        laplacian[faces[:, 0], faces[:, 1]] = -1
        laplacian[faces[:, 1], faces[:, 0]] = -1
        laplacian[faces[:, 1], faces[:, 2]] = -1
        laplacian[faces[:, 2], faces[:, 1]] = -1
        laplacian[faces[:, 2], faces[:, 0]] = -1
        laplacian[faces[:, 0], faces[:, 2]] = -1

        r, c = np.diag_indices(laplacian.shape[0])
        laplacian[r, c] = -laplacian.sum(1)

        for i in range(self.nv):
            laplacian[i, :] /= laplacian[i, i]

        self.register_buffer('laplacian', torch.from_numpy(laplacian))

    def forward(self, x):
        batch_size = x.size(0)
        x = torch.matmul(self.laplacian, x)
        dims = tuple(range(x.ndimension())[1:])
        x = x.pow(2).sum(dims)
        if self.average:
            return x.sum() / batch_size
        else:
            return x


class FlattenLoss(nn.Module):
    def __init__(self, faces, average=False):
        super(FlattenLoss, self).__init__()
        self.nf = faces.size(0)
        self.average = average

        faces = faces.detach().cpu().numpy()
        vertices = list(set([tuple(v) for v in np.sort(np.concatenate((faces[:, 0:2], faces[:, 1:3]), axis=0))]))

        v0s = np.array([v[0] for v in vertices], 'int32')
        v1s = np.array([v[1] for v in vertices], 'int32')
        v2s = []
        v3s = []
        for v0, v1 in zip(v0s, v1s):
            count = 0
            for face in faces:
                if v0 in face and v1 in face:
                    v = np.copy(face)
                    v = v[v != v0]
                    v = v[v != v1]
                    if count == 0:
                        v2s.append(int(v[0]))
                        count += 1
                    else:
                        v3s.append(int(v[0]))
        v2s = np.array(v2s, 'int32')
        v3s = np.array(v3s, 'int32')

        self.register_buffer('v0s', torch.from_numpy(v0s).long())
        self.register_buffer('v1s', torch.from_numpy(v1s).long())
        self.register_buffer('v2s', torch.from_numpy(v2s).long())
        self.register_buffer('v3s', torch.from_numpy(v3s).long())

    def forward(self, vertices, eps=1e-6):
        # make v0s, v1s, v2s, v3s
        batch_size = vertices.size(0)

        v0s = vertices[:, self.v0s, :]
        v1s = vertices[:, self.v1s, :]
        v2s = vertices[:, self.v2s, :]
        v3s = vertices[:, self.v3s, :]

        a1 = v1s - v0s
        b1 = v2s - v0s
        a1l2 = a1.pow(2).sum(-1)
        b1l2 = b1.pow(2).sum(-1)
        a1l1 = (a1l2 + eps).sqrt()
        b1l1 = (b1l2 + eps).sqrt()
        ab1 = (a1 * b1).sum(-1)
        cos1 = ab1 / (a1l1 * b1l1 + eps)
        sin1 = (1 - cos1.pow(2) + eps).sqrt()
        c1 = a1 * (ab1 / (a1l2 + eps))[:, :, None]
        cb1 = b1 - c1
        cb1l1 = b1l1 * sin1

        a2 = v1s - v0s
        b2 = v3s - v0s
        a2l2 = a2.pow(2).sum(-1)
        b2l2 = b2.pow(2).sum(-1)
        a2l1 = (a2l2 + eps).sqrt()
        b2l1 = (b2l2 + eps).sqrt()
        ab2 = (a2 * b2).sum(-1)
        cos2 = ab2 / (a2l1 * b2l1 + eps)
        sin2 = (1 - cos2.pow(2) + eps).sqrt()
        c2 = a2 * (ab2 / (a2l2 + eps))[:, :, None]
        cb2 = b2 - c2
        cb2l1 = b2l1 * sin2

        cos = (cb1 * cb2).sum(-1) / (cb1l1 * cb2l1 + eps)

        dims = tuple(range(cos.ndimension())[1:])
        loss = (cos + 1).pow(2).sum(dims)
        if self.average:
            return loss.sum() / batch_size
        else:
            return loss



# Copyright (c) Felix Petersen.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import torch
import torch.nn.functional as F
import numpy as np

import gendr


class Mesh(object):
    '''
    A simple class for creating and manipulating trimesh objects
    '''
    def __init__(self, vertices, faces, textures=None, texture_res=1, texture_type='surface'):
        '''
        vertices, faces and textures(if not None) are expected to be Tensor objects
        '''
        self._vertices = vertices
        self._faces = faces

        if isinstance(self._vertices, np.ndarray):
            self._vertices = torch.from_numpy(self._vertices).float().cuda()
        if isinstance(self._faces, np.ndarray):
            self._faces = torch.from_numpy(self._faces).int().cuda()
        if self._vertices.ndimension() == 2:
            self._vertices = self._vertices[None, :, :]
        if self._faces.ndimension() == 2:
            self._faces = self._faces[None, :, :]

        self.device = self._vertices.device
        self.texture_type = texture_type

        self.batch_size = self._vertices.shape[0]
        self.num_vertices = self._vertices.shape[1]
        self.num_faces = self._faces.shape[1]

        # create textures
        if textures is None:
            if texture_type == 'surface':
                self._textures = torch.ones(self.batch_size, self.num_faces, texture_res**2, 3,
                                            dtype=torch.float32).to(self.device)
                self.texture_res = texture_res
            elif texture_type == 'vertex':
                self._textures = torch.ones(self.batch_size, self.num_vertices, 3,
                                            dtype=torch.float32).to(self.device)
                self.texture_res = 1  # vertex color doesn't need image_size
        else:
            if isinstance(textures, np.ndarray):
                textures = torch.from_numpy(textures).float().cuda()
            if textures.ndimension() == 3 and texture_type == 'surface':
                textures = textures[None, :, :, :]
            if textures.ndimension() == 2 and texture_type == 'vertex':
                textures = textures[None, :, :]
            self._textures = textures
            self.texture_res = int(np.sqrt(self._textures.shape[2]))

    @classmethod
    def from_obj(cls, filename_obj, normalization=False, load_texture=False, texture_res=1, texture_type='surface'):
        '''
        Create a Mesh object from a .obj file
        '''
        if load_texture:
            vertices, faces, textures = gendr.functional.load_obj(filename_obj,
                                                     normalization=normalization,
                                                     texture_res=texture_res,
                                                     load_texture=True,
                                                     texture_type=texture_type)
        else:
            vertices, faces = gendr.functional.load_obj(filename_obj,
                                           normalization=normalization,
                                           texture_res=texture_res,
                                           load_texture=False)
            textures = None
        return cls(vertices, faces, textures, texture_res, texture_type)

    def save_obj(self, filename_obj, save_texture=False, texture_res_out=16):
        if self.batch_size != 1:
            raise ValueError('Could not save when batch size > 1')
        if save_texture:
            gendr.functional.save_obj(filename_obj, self.vertices[0], self.faces[0],
                         textures=self.textures[0],
                         texture_res=texture_res_out, texture_type=self.texture_type)
        else:
            gendr.functional.save_obj(filename_obj, self.vertices[0], self.faces[0], textures=None)

    @property
    def faces(self):
        return self._faces

    @property
    def vertices(self):
        return self._vertices

    @property
    def textures(self):
        return self._textures

    @property
    def face_vertices(self):
        return gendr.functional.face_vertices(self.vertices, self.faces)

    @property
    def surface_normals(self):
        v10 = self.face_vertices[:, :, 0] - self.face_vertices[:, :, 1]
        v12 = self.face_vertices[:, :, 2] - self.face_vertices[:, :, 1]
        return F.normalize(torch.cross(v12, v10), p=2, dim=2, eps=1e-6)

    @property
    def vertex_normals(self):
        return gendr.functional.vertex_normals(self.vertices, self.faces)

    @property
    def face_textures(self):
        if self.texture_type in ['surface']:
            return self.textures
        elif self.texture_type in ['vertex']:
            return gendr.functional.face_vertices(self.textures, self.faces)
        else:
            raise ValueError('texture type not applicable')

    def voxelize(self, voxel_size=32):
        face_vertices_norm = self.face_vertices * voxel_size / (voxel_size - 1) + 0.5
        return gendr.functional.voxelization(face_vertices_norm, voxel_size, False)



# Copyright (c) Felix Petersen.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import os
import torch
from skimage.io import imsave

import gendr.cuda.create_texture_image as create_texture_image_cuda


def create_texture_image(textures, texture_res=16):
    num_faces = textures.shape[0]
    tile_width = int((num_faces - 1.) ** 0.5) + 1
    tile_height = int((num_faces - 1.) / tile_width) + 1
    image = torch.ones(tile_height * texture_res, tile_width * texture_res, 3, dtype=torch.float32)
    vertices = torch.zeros((num_faces, 3, 2), dtype=torch.float32)  # [:, :, UV]
    face_nums = torch.arange(num_faces)
    column = face_nums % tile_width
    row = face_nums // tile_width
    vertices[:, 0, 0] = column * texture_res + texture_res / 2
    vertices[:, 0, 1] = row * texture_res + 1
    vertices[:, 1, 0] = column * texture_res + 1
    vertices[:, 1, 1] = (row + 1) * texture_res - 1 - 1
    vertices[:, 2, 0] = (column + 1) * texture_res - 1 - 1
    vertices[:, 2, 1] = (row + 1) * texture_res - 1 - 1
    image = image.cuda()
    vertices = vertices.cuda()
    textures = textures.cuda()
    image = create_texture_image_cuda.create_texture_image(vertices, textures, image, 1e-5)

    vertices[:, :, 0] /= (image.shape[1] - 1)
    vertices[:, :, 1] /= (image.shape[0] - 1)

    image = image.detach().cpu().numpy()
    vertices = vertices.detach().cpu().numpy()
    image = image[::-1, ::1]

    return image, vertices


def save_obj(filename, vertices, faces, textures=None, texture_res=16, texture_type='surface'):
    assert vertices.ndimension() == 2
    assert faces.ndimension() == 2
    assert texture_type in ['surface', 'vertex']
    assert texture_res >= 2

    if textures is not None and texture_type == 'surface':
        filename_mtl = filename[:-4] + '.mtl'
        filename_texture = filename[:-4] + '.png'
        material_name = 'material_1'
        texture_image, vertices_textures = create_texture_image(textures, texture_res)
        texture_image = texture_image.clip(0, 1)
        texture_image = (texture_image * 255).astype('uint8')
        imsave(filename_texture, texture_image)

    faces = faces.detach().cpu().numpy()

    with open(filename, 'w') as f:
        f.write('# %s\n' % os.path.basename(filename))
        f.write('#\n')
        f.write('\n')

        if textures is not None:
            f.write('mtllib %s\n\n' % os.path.basename(filename_mtl))

        if textures is not None and texture_type == 'vertex':
            for vertex, color in zip(vertices, textures):
                f.write('v %.8f %.8f %.8f %.8f %.8f %.8f\n' % (vertex[0], vertex[1], vertex[2],
                                                               color[0], color[1], color[2]))
            f.write('\n')
        else:
            for vertex in vertices:
                f.write('v %.8f %.8f %.8f\n' % (vertex[0], vertex[1], vertex[2]))
            f.write('\n')

        if textures is not None and texture_type == 'surface':
            for vertex in vertices_textures.reshape((-1, 2)):
                f.write('vt %.8f %.8f\n' % (vertex[0], vertex[1]))
            f.write('\n')

            f.write('usemtl %s\n' % material_name)
            for i, face in enumerate(faces):
                f.write('f %d/%d %d/%d %d/%d\n' % (
                    face[0] + 1, 3 * i + 1, face[1] + 1, 3 * i + 2, face[2] + 1, 3 * i + 3))
            f.write('\n')
        else:
            for face in faces:
                f.write('f %d %d %d\n' % (face[0] + 1, face[1] + 1, face[2] + 1))

    if textures is not None and texture_type == 'surface':
        with open(filename_mtl, 'w') as f:
            f.write('newmtl %s\n' % material_name)
            f.write('map_Kd %s\n' % os.path.basename(filename_texture))


def save_voxel(filename, voxel):
    vertices = []
    for i in range(voxel.shape[0]):
        for j in range(voxel.shape[1]):
            for k in range(voxel.shape[2]):
                if voxel[i, j, k] == 1:
                    vertices.append([i / voxel.shape[0], j / voxel.shape[1], k / voxel.shape[2]])
    vertices = torch.autograd.Variable(torch.tensor(vertices))
    return save_obj(filename, vertices, torch.autograd.Variable(torch.tensor([])))



# Copyright (c) Felix Petersen.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import torch
import torch.nn as nn
import torch.nn.functional as F
import numpy as np


def ambient_lighting(light, light_intensity=0.5, light_color=(1, 1, 1)):
    device = light.device

    if isinstance(light_color, tuple) or isinstance(light_color, list):
        light_color = torch.tensor(light_color, dtype=torch.float32, device=device)
    elif isinstance(light_color, np.ndarray):
        light_color = torch.from_numpy(light_color).float().to(device)
    if light_color.ndimension() == 1:
        light_color = light_color[None, :]

    light += light_intensity * light_color[:, None, :]
    return light  # [nb, :, 3]


def directional_lighting(light, normals, light_intensity=0.5, light_color=(1, 1, 1),
                         light_direction=(0, 1, 0)):
    # normals: [nb, :, 3]

    device = light.device

    if isinstance(light_color, tuple) or isinstance(light_color, list):
        light_color = torch.tensor(light_color, dtype=torch.float32, device=device)
    elif isinstance(light_color, np.ndarray):
        light_color = torch.from_numpy(light_color).float().to(device)
    if isinstance(light_direction, tuple) or isinstance(light_direction, list):
        light_direction = torch.tensor(light_direction, dtype=torch.float32, device=device)
    elif isinstance(light_direction, np.ndarray):
        light_direction = torch.from_numpy(light_direction).float().to(device)
    if light_color.ndimension() == 1:
        light_color = light_color[None, :]
    if light_direction.ndimension() == 1:
        light_direction = light_direction[None, :]  # [nb, 3]

    cosine = F.relu(torch.sum(normals * light_direction, dim=2))  # []
    light += light_intensity * (light_color[:, None, :] * cosine[:, :, None])
    return light  # [nb, :, 3]




# Copyright (c) Felix Petersen.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from setuptools import setup

from torch.utils.cpp_extension import BuildExtension, CUDAExtension

CUDA_FLAGS = []

ext_modules = [
    CUDAExtension('gendr.cuda.load_textures', [
        'gendr/cuda/load_textures_cuda.cpp',
        'gendr/cuda/load_textures_cuda_kernel.cu',
        ]),
    CUDAExtension('gendr.cuda.create_texture_image', [
        'gendr/cuda/create_texture_image_cuda.cpp',
        'gendr/cuda/create_texture_image_cuda_kernel.cu',
        ]),
    CUDAExtension('gendr.cuda.generalized_renderer', [
        'gendr/cuda/generalized_renderer_cuda.cpp',
        'gendr/cuda/generalized_renderer_cuda_kernel.cu',
        ]),
    CUDAExtension('gendr.cuda.voxelization', [
        'gendr/cuda/voxelization_cuda.cpp',
        'gendr/cuda/voxelization_cuda_kernel.cu',
        ]),
    ]

INSTALL_REQUIREMENTS = ['numpy', 'torch>=1.9.0', 'scikit-image', 'tqdm', 'imageio']

with open('README.md', 'r', encoding='utf-8') as fh:
    long_description = fh.read()

setup(
    name='gendr',
    version='v0.1.0',
    description='GenDR - The Generalized Differentiable Renderer',
    author='Felix Petersen',
    author_email='ads0475@felix-petersen.de',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/Felix-Petersen/gendr',
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Mathematics',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Topic :: Software Development',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    license='MIT License',
    packages=['gendr', 'gendr.cuda', 'gendr.functional'],
    install_requires=INSTALL_REQUIREMENTS,
    ext_modules=ext_modules,
    cmdclass={'build_ext': BuildExtension}
)



# Copyright (c) Felix Petersen.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import torch.nn as nn
import torch.nn.functional as F

import gendr


class GenDR(nn.Module):
    def __init__(self,
                 image_size=256,
                 background_color=[0, 0, 0],
                 anti_aliasing=False,
                 #
                 dist_func='uniform',
                 dist_scale=1e-2,
                 dist_squared=False,
                 dist_shape=None,
                 dist_shift=None,
                 dist_eps=1e4,
                 #
                 aggr_alpha_func='probabilistic',
                 aggr_alpha_t_conorm_p=None,
                 #
                 aggr_rgb_func='softmax',
                 aggr_rgb_eps=1e-3,
                 aggr_rgb_gamma=1e-3,
                 #
                 near=1,
                 far=100,
                 double_side=False,
                 texture_type='surface',
                 ):
        super(GenDR, self).__init__()

        if aggr_rgb_func not in ['hard', 'softmax']:
            raise ValueError('Aggregate function (RGB) currently only supports hard and softmax.')
        if texture_type not in ['surface', 'vertex']:
            raise ValueError('Texture type only support surface and vertex.')

        self.image_size = image_size
        self.background_color = background_color
        self.anti_aliasing = anti_aliasing

        self.dist_func = dist_func
        self.dist_scale = dist_scale
        self.dist_squared = dist_squared
        self.dist_shape = dist_shape
        self.dist_shift = dist_shift
        self.dist_eps = dist_eps

        self.aggr_alpha_func = aggr_alpha_func
        self.aggr_alpha_t_conorm_p = aggr_alpha_t_conorm_p

        self.aggr_rgb_func = aggr_rgb_func
        self.aggr_rgb_eps = aggr_rgb_eps
        self.aggr_rgb_gamma = aggr_rgb_gamma

        self.near = near
        self.far = far
        self.double_side = double_side
        self.texture_type = texture_type

    def forward(self, mesh):
        image_size = self.image_size * (2 if self.anti_aliasing else 1)

        images = gendr.functional.render(
            face_vertices=mesh.face_vertices,
            textures=mesh.face_textures,
            image_size=image_size,
            background_color=self.background_color,
            dist_func=self.dist_func,
            dist_scale=self.dist_scale,
            dist_squared=self.dist_squared,
            dist_shape=self.dist_shape,
            dist_shift=self.dist_shift,
            dist_eps=self.dist_eps,
            aggr_alpha_func=self.aggr_alpha_func,
            aggr_alpha_t_conorm_p=self.aggr_alpha_t_conorm_p,
            aggr_rgb_func=self.aggr_rgb_func,
            aggr_rgb_eps=self.aggr_rgb_eps,
            aggr_rgb_gamma=self.aggr_rgb_gamma,
            near=self.near,
            far=self.far,
            double_side=self.double_side,
            texture_type=self.texture_type,
        )

        if self.anti_aliasing:
            images = F.avg_pool2d(images, kernel_size=2, stride=2)

        return images

    def forward_tensors(self, face_vertices, face_textures):
        image_size = self.image_size * (2 if self.anti_aliasing else 1)

        images = gendr.functional.render(
            face_vertices=face_vertices,
            textures=face_textures,
            image_size=image_size,
            background_color=self.background_color,
            dist_func=self.dist_func,
            dist_scale=self.dist_scale,
            dist_squared=self.dist_squared,
            dist_shape=self.dist_shape,
            dist_shift=self.dist_shift,
            dist_eps=self.dist_eps,
            aggr_alpha_func=self.aggr_alpha_func,
            aggr_alpha_t_conorm_p=self.aggr_alpha_t_conorm_p,
            aggr_rgb_func=self.aggr_rgb_func,
            aggr_rgb_eps=self.aggr_rgb_eps,
            aggr_rgb_gamma=self.aggr_rgb_gamma,
            near=self.near,
            far=self.far,
            double_side=self.double_side,
            texture_type=self.texture_type,
        )

        if self.anti_aliasing:
            images = F.avg_pool2d(images, kernel_size=2, stride=2)

        return images



# Copyright (c) Felix Petersen.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import torch
from torch.autograd import Function
import gendr.cuda.generalized_renderer as generalized_renderer_cuda


class GenDRFunction(Function):
    @staticmethod
    def forward(
            ctx,
            face_vertices,
            textures,
            #
            image_size=256,
            background_color=[0, 0, 0],
            #
            dist_func='uniform',
            dist_scale=1e-2,
            dist_squared=False,
            dist_shape=None,
            dist_shift=None,
            dist_eps=1e4,
            #
            aggr_alpha_func='probabilistic',
            aggr_alpha_t_conorm_p=None,
            #
            aggr_rgb_func='softmax',
            aggr_rgb_eps=1e-3,
            aggr_rgb_gamma=1e-3,
            #
            near=1,
            far=100,
            double_side=True,
            texture_type='surface',
    ):

        # face_vertices: [nb, nf, 9]
        # textures: [nb, nf, 9]

        dist_func_map = {
            'hard': 0, 'heaviside': 0,
            'uniform': 1,
            'cubic_hermite': 2,
            'wigner_semicircle': 3,
            'gaussian': 4,
            'laplace': 5,
            'logistic': 6,
            'gudermannian': 7, 'hyperbolic_secant': 7,
            'cauchy': 8,
            'reciprocal': 9,
            'gumbel_max': 10,
            'gumbel_min': 11,
            'exponential': 12,
            'exponential_rev': 13,
            'gamma': 14,
            'gamma_rev': 15,
            'levy': 16,
            'levy_rev': 17,
        }
        aggr_rgb_func_map = {
            'hard': 0,
            'softmax': 1,
        }
        aggr_alpha_func_map = {
            'hard': 0,
            'max': 1,
            'probabilistic': 2,
            'einstein': 3,
            'hamacher': 4,
            'frank': 5,
            'yager': 6,
            'aczel_alsina': 7,
            'dombi': 8,
            'schweizer_sklar': 9,
        }
        texture_type_map = {
            'surface': 0,
            'vertex': 1,
        }

        # Core rendering parameters
        ctx.image_size = image_size
        ctx.background_color = background_color

        ################################################################################################################
        # Distributions
        if isinstance(dist_func, int):
            ctx.dist_func = dist_func
        else:
            ctx.dist_func = dist_func_map[dist_func]
        ctx.dist_scale = dist_scale  # tau in the paper
        assert dist_scale >= 0, dist_scale  # a negative scale is invalid
        ctx.dist_squared = dist_squared  # whether the squared input / the square-root distribution is used
        ctx.dist_shape = dist_shape  # shape parameter of the distribution
        ctx.dist_shift = dist_shift  # for shifting the distribution
        ctx.dist_eps = dist_eps  # pixels further away than dist_scale*dist_eps are ignored for performance reasons
        assert dist_eps >= 1, dist_eps  # ignoring too close to the edge makes no sense
        ################################################################################################################

        ################################################################################################################
        # T-conorms
        if isinstance(aggr_alpha_func, int):
            ctx.aggr_alpha_func = aggr_alpha_func
        else:
            ctx.aggr_alpha_func = aggr_alpha_func_map[aggr_alpha_func]
        # the shape parameter p of the t-conorm (assuming the chosen one uses this param)
        ctx.aggr_alpha_t_conorm_p = aggr_alpha_t_conorm_p
        ################################################################################################################

        ################################################################################################################
        # Shading (either hard, i.e., the closest, or using softmax as proposed in Pix2Vex and SoftRas)
        if isinstance(aggr_rgb_func, int):
            ctx.aggr_rgb_func = aggr_rgb_func
        else:
            ctx.aggr_rgb_func = aggr_rgb_func_map[aggr_rgb_func]
        ctx.aggr_rgb_eps = aggr_rgb_eps
        ctx.aggr_rgb_gamma = aggr_rgb_gamma
        ################################################################################################################

        # Some additional rendering parameters
        ctx.near = near
        ctx.far = far
        ctx.double_side = double_side  # render both sides of the triangle
        ctx.texture_type = texture_type_map[texture_type]

        face_vertices = face_vertices.clone()
        textures = textures.clone()

        ctx.device = face_vertices.device
        ctx.batch_size, ctx.num_faces = face_vertices.shape[:2]

        faces_info = torch.zeros(
            (ctx.batch_size, ctx.num_faces, 9*3),
            dtype=torch.float32,
            device=ctx.device)  # [inv*9, sym*9, obt*3, 0*6]
        aggrs_info = torch.zeros(
            (ctx.batch_size, 2, ctx.image_size, ctx.image_size),
            dtype=torch.float32,
            device=ctx.device)
        soft_colors = torch.ones(
            (ctx.batch_size, 4, ctx.image_size, ctx.image_size),
            dtype=torch.float32,
            device=ctx.device)

        soft_colors[:, 0, :, :] *= background_color[0]
        soft_colors[:, 1, :, :] *= background_color[1]
        soft_colors[:, 2, :, :] *= background_color[2]

        faces_info, aggrs_info, soft_colors = \
            generalized_renderer_cuda.forward_render(
                face_vertices,
                textures,
                faces_info,
                aggrs_info,
                soft_colors,
                #
                ctx.image_size,
                #
                ctx.dist_func,
                ctx.dist_scale,
                ctx.dist_squared,
                ctx.dist_shape,
                ctx.dist_shift,
                ctx.dist_eps,
                #
                ctx.aggr_alpha_func,
                ctx.aggr_alpha_t_conorm_p,
                #
                ctx.aggr_rgb_func,
                ctx.aggr_rgb_eps,
                ctx.aggr_rgb_gamma,
                #
                ctx.near,
                ctx.far,
                ctx.double_side,
                ctx.texture_type,
       )

        ctx.save_for_backward(face_vertices, textures, soft_colors, faces_info, aggrs_info)
        return soft_colors

    @staticmethod
    def backward(ctx, grad_soft_colors):

        face_vertices, textures, soft_colors, faces_info, aggrs_info = ctx.saved_tensors

        grad_faces = torch.zeros_like(face_vertices,
                                      dtype=torch.float32,
                                      device=ctx.device)
        grad_textures = torch.zeros_like(textures,
                                         dtype=torch.float32,
                                         device=ctx.device)
        grad_soft_colors = grad_soft_colors.contiguous()

        grad_faces, grad_textures = \
            generalized_renderer_cuda.backward_render(
                face_vertices,
                textures,
                soft_colors,
                faces_info,
                aggrs_info,
                grad_faces,
                grad_textures,
                grad_soft_colors,
                #
                ctx.image_size,
                #
                ctx.dist_func,
                ctx.dist_scale,
                ctx.dist_squared,
                ctx.dist_shape,
                ctx.dist_shift,
                ctx.dist_eps,
                #
                ctx.aggr_alpha_func,
                ctx.aggr_alpha_t_conorm_p,
                #
                ctx.aggr_rgb_func,
                ctx.aggr_rgb_eps,
                ctx.aggr_rgb_gamma,
                #
                ctx.near,
                ctx.far,
                ctx.double_side,
                ctx.texture_type,
        )

        return (
            grad_faces, grad_textures,
            None, None, None, None, None, None, None, None, None, None, None, None, None,
            None, None, None, None, None
        )


def render(
    face_vertices,
    textures,
    #
    image_size=256,
    background_color=[0, 0, 0],
    #
    dist_func='uniform',
    dist_scale=1e-2,
    dist_squared=False,
    dist_shape=None,
    dist_shift=None,
    dist_eps=1e4,
    #
    aggr_alpha_func='probabilistic',
    aggr_alpha_t_conorm_p=None,
    #
    aggr_rgb_func='softmax',
    aggr_rgb_eps=1e-3,
    aggr_rgb_gamma=1e-3,
    #
    near=1,
    far=100,
    double_side=True,
    texture_type='surface',
):
    if face_vertices.device == "cpu":
        raise TypeError('GenDR only supports CUDA Tensors.')

    return GenDRFunction.apply(
        face_vertices,
        textures,
        image_size,
        background_color,
        dist_func,
        dist_scale,
        dist_squared,
        dist_shape,
        dist_shift,
        dist_eps,
        aggr_alpha_func,
        aggr_alpha_t_conorm_p,
        aggr_rgb_func,
        aggr_rgb_eps,
        aggr_rgb_gamma,
        near,
        far,
        double_side,
        texture_type,
    )
